PORT = 8086

MAX_CACHE_AGE = 600  # seconds

DEBUG = True

PROJECTS_FILE = 'projects.yaml'
# REPOS_CACHE_DIR = '/var/cache/webhaak'
REPOS_CACHE_DIR = '/tmp/webhaak'

SECRETKEY = 'blah123'

LOG_LOCATION = 'webhaak.log'
LOG_BACKUP_COUNT = 10

PUSHOVER_USERKEY = 'umuy2covsiv9irfr57on43smideaat'
PUSHOVER_APPTOKEN = 'ahfjrmhhwwf4p2ct7grn4jqap5j5ho'



"""Execute that tasks initiated by the webhooks."""

import logging
import os
import subprocess
import time
from datetime import datetime, timedelta

import git
import httpx
import strictyaml
from fastapi import Request
from pydantic import DirectoryPath, FilePath, field_validator, json
from pydantic_settings import BaseSettings
from rq import get_current_job
from strictyaml import Bool, Map, MapPattern, Optional, Seq, Str


class Settings(BaseSettings):
    """Configuration needed for webhaak to do its tasks, using environment variables."""

    secretkey: str
    log_dir: DirectoryPath
    jobs_log_dir: DirectoryPath = 'jobs'
    eventlog_dir: DirectoryPath
    projects_file: FilePath

    repos_cache_dir: DirectoryPath

    pushover_userkey: str
    pushover_apptoken: str

    debug: bool = False

    @field_validator('jobs_log_dir', mode='before')
    def apply_root(cls, value, values):
        """Create the actual value for jobs_log_dir, through its validator."""
        if log_dir := values.data.get('log_dir'):
            # jobs_log_dir is a subdirectory of log_dir
            return log_dir / value
        # should only happen when there was an error with log_dir
        return value


# Read the settings from the environment, based on the above configuration
settings = Settings()
print(settings.model_dump())

logger = logging.getLogger('worker')

logger.setLevel(logging.DEBUG)
fh = logging.FileHandler(os.path.join(settings.log_dir, 'webhaak.log'))
fh.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
fh.setFormatter(formatter)
logger.addHandler(fh)

# strictyaml schema for project settings
schema = MapPattern(
    Str(),
    Map(
        {
            # Key for a group of triggers
            'app_key': Str(),
            'triggers': MapPattern(
                Str(),
                Map(
                    {
                        # Key for a specific trigger
                        'trigger_key': Str(),
                        Optional('notify'): Bool(),
                        Optional('notify_on_error'): Bool(),
                        # Git repository URI
                        Optional('repo'): Str(),
                        # Parent directory for repository to clone to; defaults to REPOS_CACHE_DIR
                        Optional('repo_parent'): Str(),
                        # Only act when incoming call is about this specific branch; if unspecified, trigger will
                        # always fire
                        Optional('branch'): Str(),
                        # Execute this command (after Git pull if repo specified)
                        Optional('command'): Str(),
                        # Git author username -> friendly name mapping
                        Optional('authors'): MapPattern(Str(), Str()),
                        # Call a remote endpoint
                        Optional('call_url'): Map(
                            {
                                'url': Str(),
                                # Contains json payload
                                Optional('json', default=False): Bool(),
                                # Should POST instead of GET
                                Optional('post', default=False): Bool(),
                            }
                        ),
                        # Telegram
                        Optional('telegram_chat_id'): Str(),
                        Optional('telegram_token'): Str(),
                        # Sentry
                        Optional('ignore'): Seq(Str()),
                    }
                ),
            ),
        }
    ),
)

# Load the configuration of the various projects/hooks
with open(settings.projects_file, 'r', encoding='utf-8') as pf:
    projects = strictyaml.load(pf.read(), schema).data


async def call_url(request: Request, config):
    """Call a URL with a payload.

    :param Request request: request object from FastAPI call
    :param tuple config: configuration for this webhook
    :return: string with 'OK' or 'ERROR', and the response body in JSON or plaintext, depending on server answer
    :rtype: str, dict/str
    """
    requests_client = request.app.requests_client
    url = config[1]['call_url']['url']
    logger.info(f'Calling URL {url}')
    if config[1]['call_url']['json']:
        payload = await request.json()
    else:
        payload = await request.body()
    try:
        if config[1]['call_url']['post']:
            response = await requests_client.post(url, data=payload, headers={'User-Agent': 'webhaak'})
        else:
            response = await requests_client.get(url)
    except (httpx.ConnectError, httpx.ReadTimeout) as e:
        return 'ERROR', {'error': e}

    result = 'OK'
    if response.status_code > 200:
        result = 'ERROR'
    try:
        response_body = response.json()
    except json.decoder.JSONDecodeError:
        response_body = response.text()
    return result, response_body


def send_outgoing_webhook(config, payload):
    """Send a message through POST or GET to an external URL.

    :param tuple config: configuration for this webhook
    :param dict|str payload: message to send, in json (Python dict) or string
    """
    url = config[1]['call_url']['url']
    logger.info(f'Calling URL {url}')
    try:
        if config[1]['call_url']['post']:
            response = httpx.post(url, data=payload, headers={'User-Agent': 'webhaak'}, timeout=60)
        else:
            # This could be more useful with some URL-encoded content for example
            response = httpx.get(url)
    except (httpx.ConnectError, httpx.ReadTimeout) as e:
        return 'ERROR', {'error': e}

    result = 'OK'
    if response.status_code > 200:
        result = 'ERROR'
    try:
        response_body = response.json()
    except json.decoder.JSONDecodeError:
        response_body = response.text
    return result, response_body


def format_and_send_pushover_message(user_key, app_token, text, **kwargs):
    """Send a message through PushOver.

    It is possible to specify additional properties of the message by passing keyword
    arguments. The list of valid keywords is ``title, priority, sound,
    callback, timestamp, url, url_title, device, retry, expire and html``
    which are described in the Pushover API documentation.

    For convenience, you can simply set ``timestamp=True`` to set the
    timestamp to the current timestamp.

    :param str user_key: user key in PushOver
    :param str app_token: app token for PushOver
    :param str text: message to send
    """
    message_keywords = [
        'title',
        'priority',
        'sound',
        'callback',
        'timestamp',
        'url',
        'url_title',
        'device',
        'retry',
        'expire',
        'html',
        'attachment',
    ]
    payload = {'message': text, 'user': user_key, 'token': app_token}
    for key, value in kwargs.items():
        if key not in message_keywords:
            raise ValueError(f'{key}: invalid message parameter')
        if key == 'timestamp' and value is True:
            payload[key] = int(time.time())
        else:
            payload[key] = value
    return send_pushover_message(payload)


def send_pushover_message(payload):
    """Send a message through PushOver.

    :param dict payload: key, token and message to send
    """
    response = httpx.post(
        'https://api.pushover.net/1/messages.json', data=payload, headers={'User-Agent': 'Python'}, timeout=60
    )
    return response


def make_freshping_message(hook_info):
    """Format Freshping message based on incoming hook info.

    :param dict hook_info: information about the incoming webhook payload
    :return: Markdown formatted message with the details of the Sentry issue
    :rtype: str
    """
    if hook_info.get('response_summary') == 'Available':
        state = '✅'
    else:
        state = '🚨'
    title = f'{state} [{hook_info["check_name"]}] {hook_info["response_state"]}'

    message = hook_info.get('text', '[unknown check]')
    message = f'{message}\n→ {hook_info["response_summary"]}\n\n🔗 {hook_info["check_url"]}'

    return f'{title}\n\n{message}'


def make_sentry_message(config, hook_info):
    """Format Sentry message based on incoming hook info.

    :param tuple config: configuration for this webhook
    :param dict hook_info: information about the incoming webhook payload
    :return: Markdown formatted message with the details of the Sentry issue
    :rtype: str
    """
    filter_items = config[1].get('ignore', [])
    for filter_item in filter_items:
        if hook_info['title'] in filter_item:
            # We can skip this one
            return

    title = f'💣 [{hook_info["project_name"]}] {hook_info["title"]}'
    url = hook_info['url'].replace('?referrer=webhooks_plugin', '')

    message = f'in `{hook_info["culprit"]}`'

    if hook_info.get('message'):
        message = f'{message}\n\n{hook_info["message"]}'

    if hook_info.get('stacktrace') and hook_info.get('stacktrace') != 'Not available':
        # stacktrace = hook_info['stacktrace'].replace('\\n', '\n')
        stacktrace = hook_info['stacktrace']
        message = f'{message}\n\n```python\n{stacktrace}\n```'

    message = f'{title}\n\n{message}\n\n[{url}]({url})'

    return message


def make_statuspage_message(hook_info):
    """Format Statuspage message based on incoming hook info.

    :param dict hook_info: information about the incoming webhook payload
    :return: Markdown formatted message with the details of the Sentry issue
    :rtype: str
    """
    title = f'⚠️ {hook_info["title"]}'
    url = hook_info['url']

    message = f'Impact: {hook_info["impact"]}'
    message = f'{message}\nStatus: {hook_info["status"]}'
    message = f'{message}\n\nStarted: {hook_info["created_at"]}\nUpdated: {hook_info["updated_at"]}'

    for update in hook_info['incident_updates']:
        message = f'{message}\n\nStatus: {update["status"]}\n{update["display_at"]}\n{update["body"]}'

    message = f'{message}\n\n{url}'

    return title, message


def notify_user(result, config):
    """Send a PushOver message if configured, after git operation and command have run.

    Optionally send a message to the configured Telegram chat instead.

    result is a dictionary with fields:
      command_result
      status: 'OK' | 'error'
      type: 'command_error'
      message
    """
    try:
        trigger_config = config[1]
        projectname = f'{config[0]}>{trigger_config["title"]}'
        branch = trigger_config.get('branch', 'master')
        command = trigger_config.get('command', 'n/a')
        repo = trigger_config.get('repo', 'n/a')
        message = f'repo: {repo}\nbranch: {branch}\ncommand: {command}\nruntime: {result.get("runtime")}'
        if result.get('status') == 'OK':
            title = f'Hook for {projectname} ran successfully'
        else:
            title = f'Hook for {projectname} failed: {result.get("type")}'
            message = f'{message}\n\n{result.get("message")}'
        logging.debug(message)
        logging.info('Sending notification...')
        if trigger_config.get('telegram_chat_id') and trigger_config.get('telegram_token'):
            telegram_chat_id = trigger_config['telegram_chat_id']
            telegram_token = trigger_config['telegram_token']
            # Send to Telegram chat
            # params = {'chat_id': telegram_chat_id, 'text': make_sentry_message(config)}
            params = {'chat_id': telegram_chat_id, 'text': 'Imagine a Sentry message here. Not implemented, sorry'}
            with httpx.get(
                f'https://api.telegram.org/bot{telegram_token}/sendMessage', params=params, timeout=30
            ) as response:
                logging.info('Telegram notification sent, result was %s', str(response.status_code))
        else:
            # Use the Pushover default
            response = format_and_send_pushover_message(
                settings.pushover_userkey,
                settings.pushover_apptoken,
                message,
                title=title,
                url=result['job_url'],
                url_title='Job results',
            )
            if not response.status_code == 200:
                logging.error(response.text)
        logging.info('Notification sent')
    except AttributeError:
        logging.warning('Notification through PushOver failed because of missing configuration')


def get_trigger_settings(app_key, trigger_key):
    """Look up the trigger and return the repo and command to be updated and fired.

    :param app_key: application key part of the url
    :param trigger_key: trigger key part of the url, sub part of the config
    :return: tuple with project info and the trigger config
    """
    for project in projects:
        if projects[project]['app_key'] == app_key:
            for trigger in projects[project]['triggers']:
                if projects[project]['triggers'][trigger]['trigger_key'] == trigger_key:
                    trigger_config = projects[project]['triggers'][trigger]
                    trigger_config['title'] = trigger
                    return project, trigger_config
    return None


def get_repo_basename(repo_url):
    """Extract repository basename from its url, as that will be the name of directory it will be cloned into."""
    result = os.path.basename(repo_url)
    filename, file_extension = os.path.splitext(result)
    if file_extension == '.git':
        # Strip the .git from the name, as Git will do the same on non-bare checkouts
        result = filename
    return result


def get_repo_version(repo_dir):
    """Get version of Git repo, based on latest tag, number of commits since, and latest commit hash.

    :param repo_dir: path to the Git repository
    :return: string with version
    """
    # Make sure the working directory is our project
    try:
        version = subprocess.check_output(['git', 'describe', '--always', '--tags'], stderr=None, cwd=repo_dir).strip()
    except subprocess.CalledProcessError:
        version = ''

    try:
        # byte string needs to be converted to a string
        version = version.decode('utf-8')
    except AttributeError:
        # version already was a str
        pass
    return version


def fetch_info_to_str(fetch_info):
    """git.remote.FetchInfo to human-readable representation."""
    result = fetch_info[0].note
    return result


def update_repo(config):
    """Update (pull) the Git repo.

    :param tuple config: configuration for this webhook
    :return: result of the repo update
    :rtype: str
    """
    projectname = config[0]
    trigger_config = config[1]

    repo_url = trigger_config['repo']
    repo_parent = settings.repos_cache_dir
    if 'repo_parent' in trigger_config and trigger_config['repo_parent']:
        repo_parent = trigger_config['repo_parent']

    logger.info('[%s] Updating %s', projectname, repo_url)
    logger.info('[%s] Repo parent %s', projectname, repo_parent)

    # Ensure cache dir for webhaak exists and is writable
    if not os.path.exists(repo_parent):
        os.makedirs(repo_parent)  # throws OSError if repo_parent is not writable

    # TODO: check whether dir exists with different repository
    repo_dir = os.path.join(repo_parent, get_repo_basename(repo_url))
    logger.info('[%s] Repo dir %s', projectname, repo_dir)
    if os.path.isdir(repo_dir):
        # Repo already exists locally, do a pull
        logger.info('[%s] Repo exists, pull', projectname)

        app_repo = git.Repo(repo_dir)
        origin = app_repo.remote('origin')
        result = fetch_info_to_str(origin.fetch())  # assure we actually have data. fetch() returns useful information
        logger.info('[%s] Fetch result: %s', projectname, result)
    else:
        # Repo needs to be cloned
        logger.info('[%s] Repo does not exist yet, clone', projectname)
        app_repo = git.Repo.init(repo_dir)
        origin = app_repo.create_remote('origin', repo_url)
        origin.fetch()  # assure we actually have data. fetch() returns useful information
        # Set up a local tracking branch of a remote branch
        app_repo.create_head('master', origin.refs.master).set_tracking_branch(origin.refs.master)
    branch = 'master'
    if 'branch' in trigger_config:
        branch = trigger_config['branch']
    logger.info("[%s] checkout() branch '%s'", projectname, branch)
    result = str(app_repo.git.checkout(branch))
    # pull (so really update) the checked out branch to latest commit
    origin.pull()
    logger.info("[%s] Done pulling branch '%s'", projectname, branch)
    return result


def run_command(config, hook_info):
    """Run the command(s) defined for this trigger.

    :param tuple config: configuration for this webhook
    :param dict hook_info: information about the incoming webhook payload
    """
    projectname = config[0]
    trigger_config = config[1]
    if 'command' not in trigger_config:
        # No command to execute, return
        logger.info('[%s] No command to execute', projectname)
        return None
    command = trigger_config['command']
    # Replace some placeholders to be used in executing scripts from one of the repos
    repo_parent = settings.repos_cache_dir
    if 'repo_parent' in trigger_config and trigger_config['repo_parent']:
        repo_parent = trigger_config['repo_parent']
    if 'repo' in trigger_config:
        repo_url = trigger_config['repo']
        command = command.replace('REPODIR', os.path.join(repo_parent, get_repo_basename(repo_url)))
    command = command.replace('CACHEDIR', str(settings.repos_cache_dir))
    if 'REPOVERSION' in command:
        version = get_repo_version(os.path.join(repo_parent, projectname))
        command = command.replace('REPOVERSION', version)

    for key in hook_info:
        if isinstance(hook_info[key], str):
            command = command.replace(key.upper(), hook_info[key].replace('"', '\\"'))

    command = command.strip()  # ensure no weird line feeds and superfluous whitespace are there
    logger.info('[%s] Executing `%s`', projectname, command)

    result = subprocess.run(command, capture_output=True, check=True, shell=True, universal_newlines=True)
    return result


def do_pull_andor_command(config, hook_info):
    """Asynchronous RQ task, performing the git pulling and the specified scripting inside a subprocess.

    :param tuple config: configuration for this webhook
    :param dict hook_info: information about the incoming webhook payload
    """
    this_job = get_current_job()

    projectname = config[0]
    start_time = datetime.now()
    result = {
        'application': projectname,
        'result': 'unknown',
        'trigger': config[1],
        'job_url': this_job.meta.get('job_url'),
        'runtime': timedelta(seconds=0),  # Initialise on a run time of 0
    }
    if 'repo' in config[1]:
        if config[1].get('branch') and hook_info.get('branch') and config[1]['branch'] != hook_info['branch']:
            # Push was not on the required branch, skipping execution
            logger.info(
                '[%s] skipped updating repo as push for trigger %s was on branch: %s (config wants: %s)',
                projectname,
                config[1]['title'],
                hook_info['branch'],
                config[1]['branch'],
            )
            return
        try:
            result['repo_result'] = update_repo(config)
            logger.info('[%s] result repo: %s', projectname, str(result['repo_result']))
            result['status'] = 'OK'
        except git.GitCommandError as e:
            result = {'status': 'error', 'type': 'git_error', 'message': str(e)}
            logger.error('[%s] git_error: %s', projectname, str(e))
            result['runtime'] = datetime.now() - start_time
            notify_user(result, config)
            return
        except (OSError, KeyError) as e:
            result = {'status': 'error', 'type': 'os_error', 'message': str(e)}
            logger.error('[%s] os_error: %s', projectname, str(e))
            result['runtime'] = datetime.now() - start_time
            notify_user(result, config)
            return

    if 'command' in config[1]:
        cmd_error = None
        try:
            cmd_result = run_command(config, hook_info)
        except subprocess.CalledProcessError as e:
            logger.error('[%s] Error while executing command: %s', projectname, str(e))
            cmd_result = e.stdout
            cmd_error = f'{e}\n\n{e.stderr}'

        with open(os.path.join(settings.jobs_log_dir, f'{this_job.id}.log'), 'a', encoding='utf-8') as outfile:
            # Save output of the command ran by the job to its log
            if not isinstance(cmd_result, str):
                outfile.write(f'\n== Command returncode: {cmd_result.returncode} ======\n')
                outfile.write('== Command output ======\n')
                outfile.write(cmd_result.stdout)
                outfile.write('== Command error, if any ======\n')
                outfile.write(cmd_result.stderr)
            else:
                outfile.write('== Command output ======\n')
                outfile.write(cmd_result)
                outfile.write('== Command error ======\n')
                outfile.write(cmd_error)

        if cmd_result and cmd_result.returncode == 0:
            logger.info('[%s] success for command: %s', projectname, str(cmd_result.stdout))
            result['status'] = 'OK'
        elif not cmd_result and not cmd_error:
            logger.info('[%s] no command configured', projectname)
            result['status'] = 'OK'
        else:
            result['status'] = 'error'
            result['type'] = 'command_error'
            if cmd_result:
                result['message'] = cmd_result.stderr.strip()
                logger.error(
                    '[%s] command_error with returncode %s: %s',
                    projectname,
                    str(cmd_result.returncode),
                    cmd_result.stderr,
                )
                logger.error('[%s] stdout: %s', projectname, cmd_result.stdout)
                logger.error('[%s] stderr: %s', projectname, cmd_result.stderr)
            elif cmd_error:
                result['message'] = cmd_error.strip()
            else:
                result['message'] = 'Unknown error'

    result['runtime'] = datetime.now() - start_time

    if ('notify' not in config[1] or config[1]['notify']) or (
        result['status'] == 'error' and ('notify_on_error' in config[1] and config[1]['notify_on_error'])
    ):
        notify_user(result, config)


def do_handle_freshping(config, hook_info):
    """Assemble information about the Freshping monitoring item that was pushed.

    :param tuple config: configuration for this webhook
    :param dict hook_info: information about the incoming webhook payload
    :return: result and response of the call
    :rtype: str, dict
    """
    message = make_freshping_message(hook_info)
    if not message:
        # We can skip this one
        return
    if config[1].get('call_url'):
        # send through webhook (outgoing URL call)
        return send_outgoing_webhook(config, payload=message)
    elif config[1].get('telegram_chat_id'):
        # send Telegram message
        pass


def do_handle_inoreader_rss_message(config, hook_info):
    """Send the RSS item that was pushed on through an outgoing URL call.

    :param tuple config: configuration for this webhook
    :param dict hook_info: information about the incoming webhook payload
    :return: result and response of the call
    :rtype: str, dict
    """
    return send_outgoing_webhook(config, payload=hook_info)


def do_handle_sentry_message(config, hook_info):
    """Assemble information about the Sentry item that was pushed.

    :param tuple config: configuration for this webhook
    :param dict hook_info: information about the incoming webhook payload
    :return: result and response of the call
    :rtype: str, dict
    """
    message = make_sentry_message(config, hook_info)
    if not message:
        # We can skip this one
        return
    if config[1].get('call_url'):
        # send through webhook (outgoing URL call)
        return send_outgoing_webhook(config, payload=message)
    elif config[1].get('telegram_chat_id'):
        # send Telegram message
        pass


def do_handle_statuspage_message(config, hook_info):
    """Assemble information about the Statuspage item that was pushed.

    :param tuple config: configuration for this webhook
    :param dict hook_info: information about the incoming webhook payload
    :return: result and response of the call
    :rtype: str, dict
    """
    title, message = make_statuspage_message(hook_info)
    if not message:
        # We can skip this one
        return
    response = format_and_send_pushover_message(
        settings.pushover_userkey, settings.pushover_apptoken, message, title=title
    )
    status = 'OK'
    if not response.status_code == 200:
        status = 'ERROR'
        logging.error(response.text)
    logging.info('Notification sent')
    return status, response.text



"""Main Webhaak application (runtime)."""

import binascii
import json
import logging
import os
from contextlib import asynccontextmanager
from datetime import datetime

import httpx
from fastapi import Depends, FastAPI, HTTPException, Request
from fastapi.middleware.cors import CORSMiddleware
from redis import Redis
from rq import Queue

from webhaak import incoming, tasks

settings = tasks.settings


@asynccontextmanager
async def lifespan(the_app: FastAPI):
    """Upon start, initialise an AsyncClient and assign it to an attribute named requests_client on the app object."""
    the_app.requests_client = httpx.AsyncClient()
    yield
    await the_app.requests_client.aclose()


app = FastAPI(lifespan=lifespan)

logger = logging.getLogger('webhaak')
if settings.debug:
    logger.setLevel(logging.DEBUG)

# CORS configuration
app.add_middleware(
    CORSMiddleware,
    allow_origins=['*'],  # Allow requests from everywhere
    allow_credentials=True,
    allow_methods=['*'],
    allow_headers=['*'],
)

# Default job runtime
DEFAULT_TIMEOUT = 10 * 60


async def verify_key(secret_key: str):
    """Verify whether this endpoint contains the secret part in its URL."""
    try:
        if secret_key != settings.secretkey:
            logger.warning('Secret key incorrect trying to list triggers')
            raise HTTPException(status_code=404, detail='Secret key not found')
    except AttributeError as exc:
        logger.warning('Secret key not found trying to list triggers')
        raise HTTPException(status_code=404, detail='Secret key not found') from exc


# == Web app endpoints ======


@app.get('/')
async def indexpage():
    """Index page, just link to the project repo."""
    logger.info('Root page requested')
    return {
        'message': 'Welcome to Webhaak. See the documentation how to setup and use webhooks: '
        'https://codeberg.org/diginaut/webhaak'
    }


@app.get('/admin/{secret_key}/list', dependencies=[Depends(verify_key)])
async def list_triggers(request: Request):
    """List the app_keys and trigger_keys available.

    :param Request request: fastAPI Request object to get headers from
    :return: json Response
    :rtype: dict
    """
    logger.info('Trigger list requested')

    server_url = request.base_url

    result = {}
    for project, project_info in tasks.projects.items():
        result[project] = {
            'title': project,
            'app_key': project_info['app_key'],
            'triggers': [],
        }
        for trigger in project_info['triggers']:
            url = f'{server_url}app/{project_info["app_key"]}/{project_info["triggers"][trigger]["trigger_key"]}'
            result[project]['triggers'].append(
                {
                    'title': trigger,
                    'trigger_key': project_info['triggers'][trigger]['trigger_key'],
                    'url': url,
                }
            )
    return {'projects': result}


@app.get('/app/{app_key}/{trigger_key}')
@app.options('/app/{app_key}/{trigger_key}')
@app.post('/app/{app_key}/{trigger_key}')
async def app_trigger(app_key: str, trigger_key: str, request: Request):
    """Fire the trigger described by the configuration under `trigger_key`.

    :param str app_key: application key part of the url
    :param str trigger_key: trigger key part of the url, sub part of the config
    :param Request request: fastAPI Request object to get headers from
    :return: json Response
    :rtype: dict
    """
    logger.info('%s on app_key: %s trigger_key: %s', request.method, app_key, trigger_key)
    config = tasks.get_trigger_settings(app_key, trigger_key)
    if config is None:
        logger.error('app_key/trigger_key combo not found')
        # return Response(json.dumps({'status': 'Error'}), status=404, mimetype='application/json')
        raise HTTPException(status_code=404, detail='Error')

    event_info = ''
    hook_info = {'event_type': 'push'}

    call_external_url = False
    rss_message = False
    sentry_message = False
    statuspage_message = False
    freshping_monitor = False

    if request.method == 'POST':
        if request.headers.get('X-Gitea-Event'):
            vcs_source = 'Gitea'
        elif request.headers.get('X-Gogs-Event'):
            vcs_source = 'Gogs'
        elif request.headers.get('X-GitHub-Event'):
            vcs_source = 'GitHub'
        elif request.headers.get('X-Event-Key'):
            # Other option is to check for User-Agent: Bitbucket-Webhooks/2.0
            vcs_source = 'BitBucket'
            # Examples: pullrequest:fulfilled pullrequest:created
            event_key = request.headers.get('X-Event-Key')
            logger.debug('BitBucket event: %s', event_key)
            if 'pullrequest:' in event_key:
                hook_info['pullrequest_status'] = request.headers.get('X-Event-Key').split(':')[1].strip()
                if hook_info['pullrequest_status'] == 'fulfilled':
                    hook_info['event_type'] = 'merge'
                elif hook_info['pullrequest_status'] == 'created':
                    hook_info['event_type'] = 'new'
        elif request.headers.get('Sentry-Trace'):
            logger.debug('Sentry webhook')
            sentry_message = True
            vcs_source = 'n/a'
        else:
            vcs_source = '<unknown>'

        hook_info['vcs_source'] = vcs_source

        try:
            payload = await request.json()
        except json.decoder.JSONDecodeError as error:
            payload = {}
            logger.error('JSON decode error: %s', error)

        logger.debug(payload)
        url = ''
        if payload:
            # Debug: dump payload to disk
            event_date = datetime.now().strftime('%Y%m%d_%H%M%S_%f')
            with open(f'{settings.eventlog_dir}/{event_date}_event.json', 'w', encoding='utf-8') as outfile:
                json.dump(payload, outfile)
            with open(f'{settings.eventlog_dir}/{event_date}_headers.json', 'w', encoding='utf-8') as outfile:
                json.dump(dict(request.headers.items()), outfile)

            if 'repository' in payload:
                if 'html_url' in payload['repository']:
                    url = payload['repository']['html_url']
                elif 'links' in payload['repository']:
                    # BitBucket
                    url = payload['repository']['links']['html']['href']
            elif payload.get('items') and payload['items'][0].get('canonical'):
                hook_info['event_type'] = 'news_item'
                rss_message = True
            elif payload.get('incident'):
                hook_info['event_type'] = 'statuspage_update'
                statuspage_message = True
            elif payload.get('check_url'):
                hook_info['event_type'] = 'freshping_monitor'
                freshping_monitor = True
        # Likely some ping was sent, check if so
        if request.headers.get('X-GitHub-Event') == 'ping':
            logger.info('received %s ping for %s hook: %s ', vcs_source, payload['repository']['full_name'], url)
            return json.dumps({'msg': 'Hi!'})
        if (
            request.headers.get('X-GitHub-Event') == 'push'
            or request.headers.get('X-Gitea-Event') == 'push'
            or request.headers.get('X-Gogs-Event') == 'push'
            or request.headers.get('X-Event-Key') == 'repo:push'
        ):
            event_info = f'received push from {vcs_source} for '
        elif rss_message:
            event_info = 'received RSS item for '
        elif sentry_message:
            event_info = 'received push from Sentry for '
        elif statuspage_message:
            event_info = 'received statuspage update for '
        elif freshping_monitor:
            event_info = 'received Freshping monitoring update for '
        elif 'call_url' in config[1]:
            call_external_url = True
            event_info = 'received external URL for '
        else:
            logger.info(
                'received wrong event type from %s for %s hook: %s',
                vcs_source,
                payload.get('repository', {}).get('full_name', 'not available'),
                url,
            )
            return {'error': 'wrong event type'}

        if call_external_url:
            # Event type is 'call another URL'
            status, response = await tasks.call_url(request, config)
            return {
                'status': status,
                'message': 'Command accepted and was passed on',
                'response': response,
            }

        if not payload:
            logger.error(
                '%s unknown, as no json was received. Check that %s webhook content type is application/json',
                str(event_info),
                vcs_source,
            )

        if rss_message:
            event_info = await incoming.handle_inoreader_rss_item(payload, hook_info, event_info)
        elif sentry_message:
            await incoming.handle_sentry_message(payload, hook_info, event_info)
            status, response = tasks.do_handle_sentry_message(config, hook_info)
            return {
                'status': status,
                'message': 'Command accepted and was passed on',
                'response': response,
            }
        elif statuspage_message:
            await incoming.handle_statuspage_update(payload, hook_info, event_info)
            status, response = tasks.do_handle_statuspage_message(config, hook_info)
            return {
                'status': status,
                'message': 'Command accepted and was passed on',
                'response': response,
            }
        elif freshping_monitor:
            event_info, hook_info = await incoming.handle_freshping(payload, hook_info, event_info)
            status, response = tasks.do_handle_freshping(config, hook_info)
            return {
                'status': status,
                'message': 'Command accepted and was passed on',
                'response': response,
            }
        else:
            event_info = await incoming.determine_task(config, payload, hook_info, event_info)

    # Create RQ job (task) for this request
    redis_conn = Redis()
    # use named queue to prevent clashes with other RQ workers
    q = Queue(connection=redis_conn, name='webhaak', default_timeout=DEFAULT_TIMEOUT)

    # Delay execution task, so it can run as its own process under RQ, synchronously
    if rss_message:
        job = q.enqueue(
            tasks.do_handle_inoreader_rss_message,
            args=(
                config,
                hook_info,
            ),
        )
    else:
        job = q.enqueue(
            tasks.do_pull_andor_command,
            args=(
                config,
                hook_info,
            ),
        )
    logger.info('Enqueued job with id: %s', job.id)

    if not os.path.isdir(settings.jobs_log_dir):
        os.makedirs(settings.jobs_log_dir)
    with open(os.path.join(settings.jobs_log_dir, f'{job.id}.log'), 'w', encoding='utf-8') as outfile:
        # Write event_info to task log
        outfile.write(event_info)

    server_url = request.base_url
    job.meta['job_url'] = f'{server_url}status/{job.id}'
    job.save_meta()
    return {
        'status': 'OK',
        'message': 'Command accepted and will be run in the background',
        'job_id': job.id,
        'url': f'{server_url}status/{job.id}',
    }


@app.get('/status/{job_id}')
async def job_status(job_id: str):
    """Show the status of job `job_id`.

    :param str job_id:
    :return: dictionary with a task `status` and a `result`, including a relevant `message` on failure
    :rtype: json
    """
    logger.info('Status requested for job %s', job_id)
    redis_conn = Redis()
    q = Queue(connection=redis_conn, name='webhaak')  # use named queue to prevent clashes with other RQ workers
    job = q.fetch_job(job_id)
    if job is None:
        response = {'status': 'unknown'}
    else:
        log_contents = ''
        job_logfile_name = os.path.normpath(os.path.join(settings.jobs_log_dir, f'{job_id}.log'))
        if os.path.isfile(job_logfile_name):
            with open(job_logfile_name, 'r', encoding='utf-8') as infile:
                log_contents = infile.readlines()
        job_result = job.latest_result()
        job_result_response = 'unknown, might still be running/waiting'
        if job_result:
            job_result_response = job_result.type.name
        response = {
            'status': job.get_status(),
            'result': job_result_response,
            'log': log_contents,
        }
        if not job_result:
            response['message'] = ''
        elif job_result == job_result.Type.SUCCESSFUL:
            response['message'] = job_result.return_value
        else:
            response['message'] = job_result.exc_string
        if response['message']:
            response['message'] = response['message'].strip().split('\n')
    return response


@app.get('/monitor/monitor.html')
@app.get('/monitor/')
@app.get('/monitor')
async def monitor():
    """Respond to monitoring ping."""
    return 'OK'


def generate_key():
    """Generate a random ascii string to be used as identifier."""
    return binascii.hexlify(os.urandom(24))


@app.get('/admin/{secret_key}/get_app_key', dependencies=[Depends(verify_key)])
async def get_app_key():
    """Generate new app_key."""
    logger.info('New key requested through get_app_key')
    return {'key': generate_key().decode('utf-8')}



# gunicorn_conf.py
from multiprocessing import cpu_count

bind = "127.0.0.1:8888"

# Worker Options
workers = cpu_count() + 1
worker_class = 'uvicorn.workers.UvicornWorker'

# Logging Options
loglevel = 'debug'
accesslog = '/var/log/webhaak/access_log'
errorlog = '/var/log/webhaak/error_log'



"""Make sense of the incoming webhook requests."""

import copy
import logging

logger = logging.getLogger('webhaak')


async def handle_bitbucket_push(payload, hook_info):
    """Handle an incoming Git push hook from BitBucket.

    :param dict payload: dictionary containing the incoming webhook payload
    :param dict hook_info: dictionary containing the webhook configuration
    """
    logger.debug('Amount of changes in this push: %d', len(payload['push']['changes']))
    hook_info['commit_before'] = None  # When a branch is created, old is null; use as default
    # Only take info from the first change item
    if payload['push']['changes'][0]['old']:
        # Info on the previous commit is available (so not a new branch)
        hook_info['commit_before'] = payload['push']['changes'][0]['old']['target']['hash']
        hook_info['branch_before'] = payload['push']['changes'][0]['old']['name']
    if payload['push']['changes'][0]['new']:
        # Info about the (merge) commit is known
        hook_info['commit_after'] = payload['push']['changes'][0]['new']['target']['hash']
        hook_info['branch'] = payload['push']['changes'][0]['new']['name']
    else:
        # Likely a 'None' merge commit, so get the info from the branch that is getting merged
        hook_info['commit_after'] = payload['push']['changes'][0]['old']['target']['hash']

    if 'links' in payload['push']['changes'][0] and 'html' in payload['push']['changes'][0]['links']:
        hook_info['compare_url'] = payload['push']['changes'][0]['links']['html']['href']
    elif (
        payload['push']['changes'][0]['old']
        and 'links' in payload['push']['changes'][0]['old']
        and 'html' in payload['push']['changes'][0]['old']['links']
    ):
        hook_info['compare_url'] = payload['push']['changes'][0]['old']['links']['html']['href']
    else:
        hook_info['compare_url'] = ''

    # Whether branch was closed; most likely after a merge
    hook_info['closed'] = payload['push']['changes'][0].get('closed', False)
    hook_info['commits'] = []
    commits = payload['push']['changes'][0].get('commits', [])
    for commit in commits:
        commit_info = {'hash': commit['hash']}
        if 'user' in commit['author']:
            if 'username' in commit['author']['user']:
                commit_info['name'] = commit['author']['user']['username']
            else:
                commit_info['name'] = commit['author']['user']['nickname']
        commit_info['email'] = commit['author']['raw']
        hook_info['commits'].append(commit_info)


async def handle_bitbucket_pullrequest(payload, hook_info):
    """Handle an incoming pull request hook from BitBucket.

    :param dict payload: dictionary containing the incoming webhook payload
    :param dict hook_info: dictionary containing the webhook configuration
    """
    if 'rendered' in payload['pullrequest']:
        hook_info['pullrequest_title'] = payload['pullrequest']['rendered']['title']['raw']
        hook_info['pullrequest_description'] = payload['pullrequest']['rendered']['description']['raw']
    if 'close_source_branch' in payload['pullrequest']:
        hook_info['pullrequest_close_source_branch'] = payload['pullrequest']['close_source_branch']
    if 'state' in payload['pullrequest']:
        if payload['pullrequest']['state'] == 'MERGED':
            hook_info['pullrequest_author'] = payload['pullrequest']['author']['display_name']
            hook_info['pullrequest_closed_by'] = payload['pullrequest']['closed_by']['display_name']
    if 'links' in payload['pullrequest'] and 'html' in payload['pullrequest']['links']:
        hook_info['pullrequest_url'] = payload['pullrequest']['links']['html']


async def handle_bitbucket_actor(payload, hook_info, event_info, config):
    """Assemble information about the author of this action.

    :param dict payload: dictionary containing the incoming webhook payload
    :param dict hook_info: dictionary containing the webhook configuration
    :param str event_info: message containing information about the event, to be used to log and as feedback to user
    :param dict config: the projects configuration
    """
    event_info += ' by ' + payload['actor']['nickname']
    if 'display_name' in payload['actor']:
        event_info += f' ({payload["actor"]["display_name"]})'
    hook_info['username'] = payload['actor']['nickname']

    logger.debug(config[1])
    if 'authors' in config[1]:
        # Look up the email address in the known authors list of the project
        for author in config[1]['authors']:
            if author.lower() == hook_info['username'].lower():
                hook_info['email'] = config[1]['authors'][author]
                break
    return event_info


async def handle_git_actor(payload, hook_info, event_info):
    """Assemble information about the author of this action.

    :param dict payload: dictionary containing the incoming webhook payload
    :param dict hook_info: dictionary containing the webhook configuration
    :param str event_info: message containing information about the event, to be used to log and as feedback to user
    """
    if hook_info['vcs_source'] in ('Gitea', 'Gogs'):
        event_info += ' by ' + payload['pusher']['username']
        hook_info['username'] = payload['pusher']['username']
        hook_info['email'] = payload['pusher']['email']
    elif hook_info['vcs_source'] == 'GitHub':
        event_info += ' by ' + payload['pusher']['name']
        hook_info['username'] = payload['pusher']['name']
        hook_info['email'] = payload['pusher']['email']
    return event_info


async def handle_freshping(payload, hook_info, event_info):
    """Assemble information about the monitoring event that Freshping sent so an appropriate notification can be sent.

    :param dict payload: dictionary containing the incoming webhook payload
    :param dict hook_info: dictionary containing the webhook configuration
    :param str event_info: message containing information about the event, to be used to log and as feedback to user
    """
    event_info += payload.get('check_name', 'unknown')
    # payload contains all the info we want and need
    hook_info = copy.deepcopy(payload)
    return event_info, hook_info


async def handle_sentry_message(payload, hook_info, event_info):
    """Assemble information about the event that Sentry sent so an appropriate notification can be sent later.

    :param dict payload: dictionary containing the incoming webhook payload
    :param dict hook_info: dictionary containing the webhook configuration
    :param str event_info: message containing information about the event, to be used to log and as feedback to user
    """
    event_info += payload['project_name']
    sentry_fields = ['project_name', 'culprit', 'url', 'message']
    for field in sentry_fields:
        if field in payload:
            hook_info[field] = payload[field]
    hook_info['stacktrace'] = 'Not available'
    if 'event' in payload and payload['event'] and 'title' in payload['event']:
        hook_info['title'] = payload['event']['title']
        stacktrace = []
        if 'exception' in payload['event']:
            # Always take the last set
            frames = payload['event']['exception']['values'][-1].get('stacktrace', {}).get('frames', [])
            for frame in frames:
                function_name = frame.get('function', 'unknown')
                at_line = ''
                if 'lineno' in frame:
                    at_line = f' at line {frame["lineno"]}'
                frame_message = f'{frame["filename"]} in {function_name}{at_line}'
                stacktrace.append(frame_message)
            # Sentry puts the items of the trace from last to first in the json, so reverse the trace
            stacktrace.reverse()
        elif 'logentry' in payload['event']:
            if 'message' in payload['event']['logentry']:
                stacktrace.append(str(payload['event']['logentry']['message']))
            if 'formatted' in payload['event']['logentry']:
                stacktrace.append(str(payload['event']['logentry']['formatted']))
        logger.debug(stacktrace)
        hook_info['stacktrace'] = '\n'.join(stacktrace)

        if not hook_info.get('message'):
            hook_info['message'] = payload['event']['metadata'].get('value', 'n/a')
        if not hook_info.get('culprit'):
            # Mention the URL it happened on instead, if available
            hook_info['culprit'] = payload['event'].get('request', {}).get('url', 'n/a')
    return event_info


async def handle_statuspage_update(payload, hook_info, event_info):
    """Assemble information about the event that Statuspage sent so an appropriate notification can be sent later.

    :param dict payload: dictionary containing the incoming webhook payload
    :param dict hook_info: dictionary containing the webhook configuration
    :param str event_info: message containing information about the event, to be used to log and as feedback to user
    """
    event_info += payload['incident']['name']
    hook_info['title'] = payload['incident']['name']
    hook_info['incident_updates'] = payload['incident'].get('incident_updates', [])
    hook_info['created_at'] = payload['incident']['created_at']
    hook_info['updated_at'] = payload['incident']['updated_at']
    hook_info['impact'] = payload['incident']['impact']
    hook_info['status'] = payload['incident']['status']
    hook_info['url'] = payload['incident']['shortlink']
    return event_info


async def handle_inoreader_rss_item(payload, hook_info, event_info):
    """Assemble information about the RSS item that was pushed.

    :param dict payload: dictionary containing the incoming webhook payload
    :param dict hook_info: dictionary containing the webhook configuration
    :param str event_info: message containing information about the event, to be used to log and as feedback to user
    """
    items = payload.get('items', [])
    # Actually, we only support one item anyway? TODO: support sending a list
    for item in items:
        hook_info['title'] = item.get('title', '[untitled]')
        hook_info['url'] = item.get('canonical', [{}])[0].get('href', 'unknown')
        hook_info['message'] = item.get('summary', {}).get('content', '[untitled]')
    return event_info


async def get_commits_info(payload, hook_info):
    """Assemble extra information about the commits.

    :param dict payload: dictionary containing the incoming webhook payload
    :param dict hook_info: dictionary containing the webhook configuration
    """
    hook_info['commits'] = []
    for commit in payload['commits']:
        commit_info = {}
        if 'sha' in commit:
            commit_info['hash'] = commit['sha']
        elif 'id' in commit:
            commit_info['hash'] = commit['id']
        if 'author' in commit:
            commit_info['name'] = commit['author']['name']
            commit_info['email'] = commit['author']['email']
        hook_info['commits'].append(commit_info)


async def determine_task(config, payload, hook_info, event_info):
    """Parse the incoming webhook information and assemble the hook_info.

    :param dict config: the projects configuration
    :param dict payload: dictionary containing the incoming webhook payload
    :param dict hook_info: dictionary containing the webhook configuration
    :param str event_info: message containing information about the event, to be used to log and as feedback to user
    """
    if 'push' in payload:
        # BitBucket, which has a completely different format
        await handle_bitbucket_push(payload, hook_info)

    if 'pullrequest' in payload:
        # BitBucket pullrequest event
        await handle_bitbucket_pullrequest(payload, hook_info)

    if 'ref' in payload:
        hook_info['ref'] = payload['ref']
        if 'heads' in payload['ref']:
            hook_info['branch'] = payload['ref'].replace('refs/heads/', '')
        elif 'tags' in payload['ref']:
            hook_info['tag'] = payload['ref'].replace('refs/tags/', '')
    if 'repository' in payload:
        event_info += payload['repository']['full_name']
        hook_info['repo_name'] = payload['repository']['full_name']
        if 'name' in payload['repository']:
            hook_info['project_name'] = payload['repository']['name']
    if 'actor' in payload:
        # BitBucket pusher; no email address known here though
        event_info = await handle_bitbucket_actor(payload, hook_info, event_info, config)
    if 'pusher' in payload:
        # GitHub, gitea, gogs
        event_info = await handle_git_actor(payload, hook_info, event_info)
    if 'compare' in payload:
        event_info += ', compare: ' + payload['compare']
        hook_info['compare_url'] = payload['compare']
    elif 'compare_url' in payload:
        # GitHub, gitea, gogs
        event_info += ', compare: ' + payload['compare_url']
        hook_info['compare_url'] = payload['compare_url']
    if 'before' in payload:
        hook_info['commit_before'] = payload['before']
    if 'after' in payload:
        hook_info['commit_after'] = payload['after']
    if 'commits' in payload:
        # Gather info on the commits included in this push
        await get_commits_info(payload, hook_info)

    logger.debug(hook_info)
    logger.info(event_info)

    return event_info



REDIS_URL = 'redis://localhost:6379/0'

# You can also specify the Redis DB to use
# REDIS_HOST = 'redis.example.com'
# REDIS_PORT = 6380
# REDIS_DB = 3
# REDIS_PASSWORD = 'very secret'

# Queues to listen on
QUEUES = ['webhaak']

# If you're using Sentry to collect your runtime exceptions, you can use this
# to configure RQ for it in a single step
# The 'sync+' prefix is required for raven: https://github.com/nvie/rq/issues/350#issuecomment-43592410
# SENTRY_DSN = 'sync+http://public:secret@example.com/1'

# If you want custom worker name
# NAME = 'worker-1024'
# NAME = 'devworker-42'
